#define _DEFAULT_SOURCE

#include <lektor/common.h>
#include <lektor/mkv.h>
#include <dirent.h>
#include <sys/wait.h>

#define METADATA_TEMPLATE                        \
    "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" \
    "<Tags>                                    " \
    "<Tag>                                     " \
    "  <Targets>                               " \
    "    <TargetTypeValue>70</TargetTypeValue> " \
    "  </Targets>                              " \
    "  <Simple>                                " \
    "    <Name>TITLE</Name>                    " \
    "    <String>%s</String>                   " \
    "  </Simple>                               " \
    "</Tag>                                    " \
    "<Tag>                                     " \
    "  <Simple>                                " \
    "    <Name>TITLE</Name>                    " \
    "    <String>%s</String>                   " \
    "  </Simple>                               " \
    "  <Simple>                                " \
    "    <Name>CONTENT_TYPE</Name>             " \
    "    <String>%s</String>                   " \
    "  </Simple>                               " \
    "  <Simple>                                " \
    "    <Name>ADDRESS</Name>                  " \
    "    <String>%s</String>                   " \
    "  </Simple>                               " \
    "  <Simple>                                " \
    "    <Name>ARTIST</Name>                   " \
    "    <String>%s</String>                   " \
    "  </Simple>                               " \
    "  <Simple>                                " \
    "    <Name>GENRE</Name>                    " \
    "    <String>%s</String>                   " \
    "  </Simple>                               " \
    "  <Simple>                                " \
    "    <Name>TRACK</Name>                    " \
    "    <String>%d</String>                   " \
    "  </Simple>                               " \
    "</Tag>                                    " \
    "</Tags>                                   "

static bool
mkvpropedit__(const char *const args[])
{
    pid_t pid;
    int wstatus, status, fd;

    if ((pid = fork()) == 0) {
        if ((fd = open("/dev/null", O_WRONLY | O_TRUNC)) < 0) {
            LOG_ERROR("FORK", "Can't to open /dev/null in O_WRONLY O_TRUNC");
            return false;
        }

        if (dup2(fd, 1) < 0) {
            LOG_ERROR("FORK", "Failed to duplicate /dev/null to stdout");
            return false;
        }

        /* Let exec search the PATH for us */
        execvp(args[0], (char *const *)args);
        exit(EXIT_FAILURE);
    }

    else if (pid < 0) {
        LOG_ERROR("FORK", "Failed to fork: %s\n", strerror(errno));
        return false;
    }

    else {
        do {
            if (waitpid(pid, &wstatus, WUNTRACED | WCONTINUED) == -1) {
                LOG_ERROR("FORK", "Failed to wait children: %s", strerror(errno));
                return false;
            }
        } while (!WIFEXITED(wstatus) && !WIFSIGNALED(wstatus));

        if ((status = WEXITSTATUS(wstatus))) {
            LOG_ERROR("FORK", "Children failed with status %d", status);
            if (status == 2)
                LOG_ERROR(
                    "FORK",
                    "Status 2 with mkvpropedit could means that karas are not valid MKV files");
            else if (status == 1)
                LOG_WARN("FORK",
                         "Status 1 on mkvpropedit means that it generated some warnings...");
            return false;
        }
    }

    return true;
}

int
kara_metadata_write(struct kara_metadata *mdt, const char *filename)
{
    char tmpfilepath[PATH_MAX];
    char *const metadafilepath = &tmpfilepath[4];
    const char *args1[]        = { "mkvpropedit", "-t", "all:", filename, NULL };
    const char *args2[]        = { "mkvpropedit", "-t", tmpfilepath, filename, NULL };
    int fd, sta = -1;

    memset(tmpfilepath, 0, PATH_MAX);
    strncat(tmpfilepath, "all:/tmp/lektor.metadata.XXXXXX", PATH_MAX - 1);

    if ((fd = mkstemp(metadafilepath)) < 0) {
        LOG_ERROR("MKV", "Failed to create temporary file: %s", strerror(errno));
        goto error;
    }

    if (dprintf(fd, METADATA_TEMPLATE, mdt->source_name, mdt->song_name, mdt->category,
                mdt->language, mdt->author_name, mdt->song_type, mdt->song_number) < 0) {
        LOG_ERROR("MKV", "Failed to write to temporary file: %s", metadafilepath);
        goto error;
    }

    if (!mkvpropedit__(args1) || !mkvpropedit__(args2))
        goto error;

    sta = 0;
error:
    close(fd);
    unlink(metadafilepath);
    return sta;
}

void
kara_metadata_rules_from_config(struct kara_metadata_rules *rules, lkt_db *db)
{
    // clang-format off
    DBCONF_GET_TEXT_NOSPACE(db, "database", "types",      rules->song_type_rules, LKT_LINE_MAX);
    DBCONF_GET_TEXT_NOSPACE(db, "database", "categories", rules->category_rules,  LKT_LINE_MAX);
    DBCONF_GET_TEXT_NOSPACE(db, "database", "languages",  rules->language_rules,  LKT_LINE_MAX);
    // clang-format on

#define CHECK_FIELD(field)                                 \
    for (size_t i = 0; i < strlen(field); ++i) {           \
        char c = (field)[i];                               \
        FAIL_UNLESS(isalpha(c) || c == ',',                \
                    "Invalid character detected in '%s', " \
                    "expected alpha characters or commas", \
                    (field));                              \
        if (c == ',')                                      \
            (field)[i] = '|';                              \
    }
    CHECK_FIELD(rules->song_type_rules);
    CHECK_FIELD(rules->category_rules);
    CHECK_FIELD(rules->language_rules);
#undef CHECK_FIELD
}

PRIVATE_FUNCTION bool
metadata_from_path(char *const mkvfile, struct kara_metadata *meta, regex_t *regex)
{
#define NMATCH 10

    regmatch_t pmatch[NMATCH];
    char msgbuf[LEKTOR_TAG_MAX];
    int sta = false;

    memset(meta, 0, sizeof(struct kara_metadata));
    memset(msgbuf, 0, LEKTOR_TAG_MAX * sizeof(char));

    const int ret = regexec(regex, mkvfile, NMATCH, pmatch, 0);
    if (!ret) {
        // clang-format off
        memcpy(meta->category,    mkvfile + pmatch[2].rm_so, (size_t)(pmatch[2].rm_eo - pmatch[2].rm_so));
        memcpy(meta->language,    mkvfile + pmatch[3].rm_so, (size_t)(pmatch[3].rm_eo - pmatch[3].rm_so));
        memcpy(meta->author_name, mkvfile + pmatch[4].rm_so, (size_t)(pmatch[4].rm_eo - pmatch[4].rm_so));
        memcpy(meta->source_name, mkvfile + pmatch[5].rm_so, (size_t)(pmatch[5].rm_eo - pmatch[5].rm_so));
        memcpy(meta->song_type,   mkvfile + pmatch[6].rm_so, (size_t)(pmatch[6].rm_eo - pmatch[6].rm_so));
        memcpy(msgbuf,            mkvfile + pmatch[7].rm_so, (size_t)(pmatch[7].rm_eo - pmatch[7].rm_so));
        memcpy(meta->song_name,   mkvfile + pmatch[8].rm_so, (size_t)(pmatch[8].rm_eo - pmatch[8].rm_so));
        // clang-format on
    } else if (REG_NOMATCH == ret) {
        LOG_ERROR("MKV", "No match for: %s", mkvfile);
        goto error;
    } else {
        regerror(ret, regex, msgbuf, sizeof(msgbuf));
        LOG_ERROR("MKV", "Failed to execute regex: %s", msgbuf);
        goto error;
    }

    meta->song_number = (int)strtol(msgbuf, NULL, 0);
    if (meta->song_number <= 0)
        meta->song_number = 1;

    sta = true;
error:
    return sta;
#undef NMATCH
}

PRIVATE_FUNCTION regex_t *
setup_regex_from_rules(char rgx[LKT_LINE_MAX], struct kara_metadata_rules *rules)
{
    regex_t *regex = NULL;

    safe_snprintf(rgx, LKT_LINE_MAX,
                  "^/(.+)/(%s)/(%s)/(.+)/(.+) - (%s)([[:digit:]]*) - (.+)\\.mkv$",
                  rules->category_rules, rules->language_rules, rules->song_type_rules);

    if (regcomp(regex, rgx, REG_EXTENDED)) {
        LOG_ERROR("MKV", "Failed to compile regex");
        if (regex)
            regfree(regex);
        return NULL;
    }

    return regex;
}

int
metadata_set_directory(const char *kara_dir, lkt_db *db)
{
    DIR *d             = NULL;
    struct dirent *dir = NULL;
    char rgx[LKT_LINE_MAX];
    char path[PATH_MAX];
    struct kara_metadata meta;
    struct kara_metadata_rules rules;

    memset(path, 0, PATH_MAX * sizeof(char));
    kara_metadata_rules_from_config(&rules, db);
    regex_t *regex = setup_regex_from_rules(rgx, &rules);
    GOTO_UNLESS(regex, "Failed to init the regex!", end_error);

    if (!(d = opendir(kara_dir))) {
        LOG_ERROR("MKV", "Failed to open directory '%s': %s", kara_dir, strerror(errno));
        return 1;
    }

    while (NULL != (dir = readdir(d))) {
        strncpy(path, kara_dir, PATH_MAX - 1);
        strncat(path, "/", PATH_MAX - 1);
        strncat(path, dir->d_name, PATH_MAX - 1);

        if (dir->d_type == DT_REG && metadata_from_path(path, &meta, regex) &&
            !kara_metadata_write(&meta, path))
            continue;
        else if (dir->d_type == DT_DIR && !STR_MATCH(dir->d_name, ".") &&
                 !STR_MATCH(dir->d_name, ".."))
            metadata_set_directory(path, db);
    }

    LOG_INFO("MKV", "Passed directory '%s'", kara_dir);
end_error:
    if (d)
        closedir(d);
    regfree(regex);
    return false;
}

int
metadata_set_file(char *karapath, lkt_db *db)
{
    struct kara_metadata meta;
    struct kara_metadata_rules rules;
    char rgx[LKT_LINE_MAX];
    bool ret = false;

    kara_metadata_rules_from_config(&rules, db);
    regex_t *regex = setup_regex_from_rules(rgx, &rules);
    GOTO_UNLESS(regex, "Failed to init the regex!", end_error);

    ret = metadata_from_path(karapath, &meta, regex) && !kara_metadata_write(&meta, karapath);

end_error:
    if (regex)
        regfree(regex);
    return !ret;
}
